/***
 * Cecilia ADL Compiler
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author:Ali Erdem Ozcan
 *
 * Contributor: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.adl.attributes;

import java.util.Map;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.AbstractLoader;
import org.objectweb.fractal.adl.Definition;
import org.objectweb.fractal.adl.Loader;
import org.objectweb.fractal.adl.Node;
import org.objectweb.fractal.adl.attributes.Attribute;
import org.objectweb.fractal.adl.attributes.Attributes;
import org.objectweb.fractal.adl.attributes.AttributesContainer;
import org.objectweb.fractal.adl.components.Component;
import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.adl.interfaces.IDLException;
import org.objectweb.fractal.adl.interfaces.IDLLoader;
import org.objectweb.fractal.adl.xml.XMLNodeFactory;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.cecilia.adl.CeciliaADLConstants;
import org.objectweb.fractal.cecilia.adl.idl.ast.Field;
import org.objectweb.fractal.cecilia.adl.idl.ast.FieldContainer;
import org.objectweb.fractal.cecilia.adl.idl.ast.IDLDefinitionContainer;
import org.objectweb.fractal.cecilia.adl.idl.ast.RecordDefinition;
import org.xml.sax.SAXException;

/**
 * A {@link org.objectweb.fractal.adl.Loader} to check {@link Attributes} nodes
 * in definitions. This loader checks that the Java attribute controller
 * interfaces specified in these nodes exist, and that the attribute names and
 * values are consistent with the methods of theses interfaces.
 */
public class AttributeLoader extends AbstractLoader {

  // --------------------------------------------------------------------------
  // Client interfaces
  // --------------------------------------------------------------------------

  /**
   * The name of the {@link XMLNodeFactory} client interface used by this
   * component.
   */
  public static final String NODE_FACTORY_ITF     = "node-factory";

  /**
   * The name of the {@link IDLLoader} client interface used by this component.
   */
  public static final String INTERFACE_LOADER_ITF = "interface-loader";

  /** The {@link XMLNodeFactory} client interface used by this component. */
  protected XMLNodeFactory   nodeFactoryItf;

  /** The {@link IDLLoader} client interface used by this component. */
  public IDLLoader           interfaceCodeLoaderItf;

  // --------------------------------------------------------------------------
  // Implementation of the Loader interface
  // --------------------------------------------------------------------------

  public Definition load(final String name, final Map<Object, Object> context)
      throws ADLException {
    final Definition d = clientLoader.load(name, context);
    checkNode(d, context);
    return d;
  }

  // --------------------------------------------------------------------------
  // Checking methods
  // --------------------------------------------------------------------------

  private void checkNode(final Object node, final Map<Object, Object> context)
      throws ADLException {
    if (node instanceof AttributesContainer) {
      checkAttributesContainer((AttributesContainer) node, context);
    }
    if (node instanceof ComponentContainer) {
      final Component[] comps = ((ComponentContainer) node).getComponents();
      for (final Component element : comps) {
        checkNode(element, context);
      }
    }
  }

  private void checkAttributesContainer(final AttributesContainer container,
      final Map<Object, Object> context) throws ADLException {
    final Attributes attrs = container.getAttributes();
    if (attrs != null) {
      final String signature = attrs.getSignature();
      if (signature == null) {
        throw new ADLException("Signature missing", (Node) attrs);
      }
      final Attribute[] assignedAttributes = attrs.getAttributes();
      for (final Attribute element : assignedAttributes) {
        final String attrName = element.getName();
        final String attrValue = element.getValue();
        if (attrName == null) {
          throw new ADLException("Attribute name missing", (Node) element);
        }
        if (attrValue == null) {
          throw new ADLException("Attribute value missing (attribute: "
              + attrName + ")", (Node) element);
        }
      }

      Node recordDefinition = ((IDLDefinitionContainer) attrs)
          .getIDLDefinition();
      if (recordDefinition == null) {
        try {
          recordDefinition = (Node) interfaceCodeLoaderItf.loadInterface(
              signature, context);
        } catch (final IDLException e) {
          throw new ADLException("Invalid Attributes signature '" + signature
              + "'", (Node) container, e);
        }
        if (!(recordDefinition instanceof RecordDefinition)) {
          throw new ADLException("Invalid Attributes signature '" + signature
              + "' : not a record", (Node) container);
        }
        // Add the record definition in the container
        ((IDLDefinitionContainer) attrs)
            .setIDLDefinition((RecordDefinition) recordDefinition);
      }
      final Field[] recordFields = ((FieldContainer) recordDefinition)
          .getFields();

      final Attribute[] declaredAttributes = new Attribute[recordFields.length];
      // The AST node will be replaced by the set of declared attribute to add
      // the non-assigned attributes respecting the required order

      // Check if there are assigned attribute names which does nott exist in
      // the record
      for (final Attribute ass : assignedAttributes) {
        final String assName = ass.getName();
        boolean found = false;
        for (final Field dec : recordFields) {
          if (dec.getName().equals(assName)) {
            found = true;
            break;
          }
        }
        if (!found)
          throw new ADLException("Attribute name '" + assName
              + "' is not found in record '" + attrs.getSignature() + "'.",
              (Node) attrs, null);
      }

      // 1. Merge the assigned attribute values into the declared one
      for (int i = 0; i < declaredAttributes.length; i++) {
        final Field field = recordFields[i];
        try {
          declaredAttributes[i] = (Attribute) nodeFactoryItf.newXMLNode(
              CeciliaADLConstants.CECILIA_ADL_DTD,
              CeciliaADLConstants.ATTRIBUTE_AST_NODE_NAME);
        } catch (final SAXException e) {
          e.printStackTrace();
          throw new InternalError(e.getMessage());
        }
        declaredAttributes[i].setName(field.getName());
        declaredAttributes[i].setValue("0");

        for (final Attribute element : assignedAttributes) {
          if (field.getName().equals(element.getName()))
            declaredAttributes[i].setValue(element.getValue());
        }
      }

      for (final Attribute element : assignedAttributes)
        attrs.removeAttribute(element);

      for (final Attribute element : declaredAttributes)
        attrs.addAttribute(element);

    }
  }

  // ------------------------------------------------------------
  // BindingController implementation methods
  // ------------------------------------------------------------

  @Override
  public void bindFc(final String s, final Object o)
      throws NoSuchInterfaceException, IllegalBindingException {

    if (s == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (s.equals(LOADER_BINDING)) {
      this.clientLoader = (Loader) o;
    } else if (s.equals(NODE_FACTORY_ITF)) {
      this.nodeFactoryItf = (XMLNodeFactory) o;
    } else if (s.equals(INTERFACE_LOADER_ITF)) {
      this.interfaceCodeLoaderItf = (IDLLoader) o;
    } else {
      throw new NoSuchInterfaceException("There is no interface named '" + s
          + "'");
    }
  }

  @Override
  public String[] listFc() {
    return new String[]{LOADER_BINDING, NODE_FACTORY_ITF, INTERFACE_LOADER_ITF};
  }

  @Override
  public Object lookupFc(final String s) throws NoSuchInterfaceException {

    if (s == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (s.equals(LOADER_BINDING)) {
      return this.clientLoader;
    } else if (s.equals(NODE_FACTORY_ITF)) {
      return this.nodeFactoryItf;
    } else if (s.equals(INTERFACE_LOADER_ITF)) {
      return this.interfaceCodeLoaderItf;
    } else {
      throw new NoSuchInterfaceException("There is no interface named '" + s
          + "'");
    }
  }

  @Override
  public void unbindFc(final String s) throws IllegalBindingException,
      NoSuchInterfaceException {

    if (s == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (s.equals(LOADER_BINDING)) {
      this.clientLoader = null;
    } else if (s.equals(NODE_FACTORY_ITF)) {
      this.nodeFactoryItf = null;
    } else if (s.equals(INTERFACE_LOADER_ITF)) {
      this.interfaceCodeLoaderItf = null;
    } else {
      throw new NoSuchInterfaceException("There is no interface named '" + s
          + "'");
    }
  }

}
