/***
 * Cecilia ADL Compiler
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.adl.compiler;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.Definition;
import org.objectweb.fractal.adl.components.Component;
import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.adl.implementations.Implementation;
import org.objectweb.fractal.adl.implementations.ImplementationContainer;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.cecilia.adl.directives.Ldflag;
import org.objectweb.fractal.cecilia.adl.directives.LdflagContainer;
import org.objectweb.fractal.task.codegen.lib.AbstractRecursiveFileCollectionProviderTask;
import org.objectweb.fractal.task.core.TaskMap;

/**
 * Adds output task which provides an executable file linked from the collection
 * of every compiled files used by the compiled component.
 */
public class LinkerOutputCompiler extends AbstractOutputCompiler
    implements
      LinkerCompilerAttributes {

  // --------------------------------------------------------------------------
  // Attribute field
  // --------------------------------------------------------------------------

  private String             fileExtension;

  private String             fileName;

  // --------------------------------------------------------------------------
  // Client interfaces.
  // --------------------------------------------------------------------------

  /**
   * the name of the {@link LinkerBuilder linker builder} client interface used
   * by this compiler.
   * 
   * @see #linkerBuilderItf
   */
  public static final String LINKER_BUILDER_ITF = "linker-builder";

  /** The {@link LinkerBuilder linker builder} used by this compiler. */
  public LinkerBuilder       linkerBuilderItf;

  // --------------------------------------------------------------------------
  // Implementation of the PrimitiveCompiler interface
  // --------------------------------------------------------------------------

  public void compile(final List<ComponentContainer> path,
      final ComponentContainer container, final TaskMap tasks,
      final Map<Object, Object> context) throws ADLException {

    try {
      tasks.getTask("output", container);
      return;
    } catch (final NoSuchElementException e) {
      // continue
    }

    final StringBuffer ldFlagBuffer = new StringBuffer();
    addLdFlags(container, ldFlagBuffer);

    final File outDir = (File) context.get("objBuildDirectory");

    String name = (String) context.get("executable-name");

    if (name == null || name.length() == 0) {

      if (container instanceof Definition)
        name = ((Definition) container).getName().replace('.', '_');
      else
        name = ((Component) container).getName().replace('.', '_');
    }

    if (fileExtension != null) name += "." + fileExtension;

    final File destFile = new File(outDir, name);

    final LinkTask outputTask = new LinkTask(linkerBuilderItf, destFile,
        ldFlagBuffer.toString());
    tasks.addTask("output", container, outputTask);

    final Collection<TaskMap.TaskHole> compilationTasks = getComponentCompilationTasks(
        container, tasks, context);
    for (final TaskMap.TaskHole taskHole : compilationTasks) {
      outputTask.addFileProviderTask(taskHole);
    }
    addSubCompDependancies(container, tasks, outputTask);
  }

  private void addLdFlags(final Object component,
      final StringBuffer ldFlagBuffer) {
    Implementation impl = null;
    if (component instanceof ImplementationContainer)
      impl = ((ImplementationContainer) component).getImplementation();

    if (impl != null && impl instanceof LdflagContainer) {
      for (final Ldflag ldFlag : ((LdflagContainer) impl).getLdflags()) {
        ldFlagBuffer.append(ldFlag.getValue()).append(' ');
      }
    }
    if (!(component instanceof ComponentContainer)) return;
    final Component[] subComponents = ((ComponentContainer) component)
        .getComponents();
    for (final Component element : subComponents) {
      addLdFlags(element, ldFlagBuffer);
    }
  }

  // --------------------------------------------------------------------------
  // Task classes
  // --------------------------------------------------------------------------

  protected final class LinkTask
      extends
        AbstractRecursiveFileCollectionProviderTask {

    private final LinkerBuilder builder;
    private final File          destFile;
    private final String        ldFlags;

    /**
     * @param builder
     * @param destFile
     * @param ldFlags
     */
    public LinkTask(final LinkerBuilder builder, final File destFile,
        final String ldFlags) {
      this.builder = builder;
      this.destFile = destFile;
      this.ldFlags = ldFlags;
    }

    public void execute(final Map<Object, Object> context) throws Exception {
      builder.link(destFile, getProvidedFiles().toArray(new File[0]), ldFlags,
          context);
      final Collection<File> c = new ArrayList<File>();
      c.add(destFile);
      setResult(c);
    }
  }

  // --------------------------------------------------------------------------
  // Implementation of the AttributeController interface
  // --------------------------------------------------------------------------

  public String getFileExtension() {
    return fileExtension;
  }

  public void setFileExtension(final String ext) {
    fileExtension = ext;
  }

  public final String getFileName() {
    return this.fileName;
  }

  public final void setFileName(final String fileName) {
    this.fileName = fileName;
  }

  // --------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // --------------------------------------------------------------------------

  @Override
  public String[] listFc() {
    final String[] superList = super.listFc();
    final String[] list = new String[superList.length + 1];
    System.arraycopy(superList, 0, list, 0, superList.length);
    list[superList.length] = LINKER_BUILDER_ITF;
    return list;
  }

  @Override
  public void bindFc(final String itf, final Object value)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException {

    if (itf == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (itf.equals(LINKER_BUILDER_ITF)) {
      this.linkerBuilderItf = (LinkerBuilder) value;
    } else {
      super.bindFc(itf, value);
    }
  }

  @Override
  public Object lookupFc(final String itf) throws NoSuchInterfaceException {
    if (itf == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (itf.equals(LINKER_BUILDER_ITF)) {
      return this.linkerBuilderItf;
    } else {
      return super.lookupFc(itf);
    }
  }

  @Override
  public void unbindFc(final String itf) throws NoSuchInterfaceException,
      IllegalBindingException, IllegalLifeCycleException {
    if (itf == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (itf.equals(LINKER_BUILDER_ITF)) {
      this.linkerBuilderItf = null;
    } else {
      super.unbindFc(itf);
    }
  }

}
