/***
 * Cecilia ADL Compiler
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author:Ali Erdem Ozcan
 */

package org.objectweb.fractal.cecilia.adl.controllers.binding;

import static org.objectweb.fractal.cecilia.adl.bindings.BindingDecorationUtil.isBindAfterStart;
import static org.objectweb.fractal.cecilia.adl.bindings.BindingDecorationUtil.isStaticallyBound;
import static org.objectweb.fractal.cecilia.adl.bindings.BindingDecorationUtil.isUsedAtStart;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.bindings.Binding;
import org.objectweb.fractal.adl.bindings.BindingContainer;
import org.objectweb.fractal.adl.components.Component;
import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.cecilia.adl.bindings.BindingBuilder;
import org.objectweb.fractal.cecilia.adl.components.AbstractPrimitiveCompiler;
import org.objectweb.fractal.task.codegen.api.InstanceProviderTask;
import org.objectweb.fractal.task.codegen.api.RequireFactoryProviderTask;
import org.objectweb.fractal.task.codegen.api.RequireInstanceProviderTask;
import org.objectweb.fractal.task.codegen.api.RequireSourceCodeProviderTask;
import org.objectweb.fractal.task.codegen.lib.AbstractRecursiveSourceCodeProviderTask;
import org.objectweb.fractal.task.core.TaskMap;
import org.objectweb.fractal.task.core.TaskMap.TaskHole;

/**
 * Compiler for bindings.
 */
public class CompositeBCCompiler extends AbstractPrimitiveCompiler
    implements
      BindingController {

  // --------------------------------------------------------------------------
  // Client interface.
  // --------------------------------------------------------------------------

  /**
   * The name of the {@link BindingBuilder} client interface used by this
   * compiler.
   */
  public static final String BUILDER_ITF = "builder";

  /** The {@link BindingBuilder} used by this compiler. */
  public BindingBuilder      builderItf;

  // --------------------------------------------------------------------------
  // Implementation of the Compiler interface
  // --------------------------------------------------------------------------

  public void compile(final List<ComponentContainer> path,
      final ComponentContainer container, final TaskMap tasks,
      final Map<Object, Object> context) throws ADLException {

    if (!(container instanceof BindingContainer)
        || getCompilerRole() != CompilerRole.instance) return;
    try {
      // the task may already exist, in case of a shared component
      tasks.getTask("bindings", container);
      return;
    } catch (final NoSuchElementException e) {
      // ignore
    }

    // special task which appregates all others
    final BindTask bindingsTask = new BindTask(builderItf, 0, null, null, null,
        null);
    final TaskMap.TaskHole bindingTaskHole = addTask("bindings", bindingsTask,
        container, tasks);
    getTaskHole("controller", container, tasks).addDependency(bindingTaskHole,
        RequireSourceCodeProviderTask.SOURCE_CODE_PROVIDER_TASK_ROLE,
        "bindings");

    // build a map of sub components
    final Map<String, ComponentContainer> subComponents = new HashMap<String, ComponentContainer>();
    subComponents.put("this", container);
    for (final Component element : container.getComponents()) {
      subComponents.put(element.getName(), element);
    }

    for (final Binding binding : ((BindingContainer) container).getBindings()) {
      final String c = binding.getFrom();
      int index = c.indexOf('.');
      final Object clientComp = subComponents.get(c.substring(0, index));
      final String clientItf = c.substring(index + 1);

      final String s = binding.getTo();
      index = s.indexOf('.');
      final Object serverComp = subComponents.get(s.substring(0, index));
      final String serverItf = s.substring(index + 1);

      final TaskMap.TaskHole instanceClientTask = tasks.getTaskHole("instance",
          clientComp);
      final TaskMap.TaskHole instanceServerTask = tasks.getTaskHole("instance",
          serverComp);

      int type = BindingBuilder.NORMAL_BINDING;
      if (binding.getFrom().startsWith("this."))
        type = BindingBuilder.EXPORT_BINDING;
      else if (binding.getTo().startsWith("this."))
        type = BindingBuilder.IMPORT_BINDING;

      if (isBindAfterStart(binding))
        type |= BindingBuilder.AFTERSTART_BINDING;
      else if (isUsedAtStart(binding))
        type |= BindingBuilder.USED_AT_START;
      else if (isStaticallyBound(binding))
        type |= BindingBuilder.STATIC_BINDING;

      final String taskName = "bind_" + clientItf;
      try {
        // the task may already exist, in case of a shared component
        tasks.getTask(getFullTaskName(taskName), clientComp);
        continue;
      } catch (final NoSuchElementException e) {
        // ignore
      }
      final BindTask bindTask = new BindTask(builderItf, type,
          instanceClientTask, clientItf, instanceServerTask, serverItf);

      final TaskHole bindTaskHole = tasks.addTask(getFullTaskName(taskName),
          clientComp, bindTask);
      ((RequireFactoryProviderTask) bindTask).setFactoryProviderTask(tasks
          .getTaskHole("type", container));
      ((RequireInstanceProviderTask) bindTask).setInstanceProviderTask(tasks
          .getTaskHole("instance", container));
      bindingsTask.addSourceCodeProviderTask("bind_" + c, bindTaskHole);
    }
  }

  // --------------------------------------------------------------------------
  // Task class
  // --------------------------------------------------------------------------

  protected static class BindTask
      extends
        AbstractRecursiveSourceCodeProviderTask {

    private final BindingBuilder   builder;
    private final int              type;
    private final TaskMap.TaskHole clientInstanceProviderTask;
    private final String           clientItf;
    private final TaskMap.TaskHole serverInstanceProviderTask;
    private final String           serverItf;

    protected BindTask(final BindingBuilder builder, final int type,
        final TaskMap.TaskHole clientTask, final String clientItf,
        final TaskMap.TaskHole serverTask, final String serverItf) {
      this.builder = builder;
      this.type = type;
      this.clientInstanceProviderTask = clientTask;
      if (clientTask != null) addPreviousTask(clientTask);
      this.clientItf = clientItf;
      this.serverInstanceProviderTask = serverTask;
      if (serverTask != null) addPreviousTask(serverTask);
      this.serverItf = serverItf;
    }

    public void execute(final Map<Object, Object> context) throws Exception {
      final Object composite = getInstanceProviderTask().getInstance();
      final Object client = (clientInstanceProviderTask == null)
          ? null
          : ((InstanceProviderTask) clientInstanceProviderTask.getTask())
              .getInstance();
      final Object server = (serverInstanceProviderTask == null)
          ? null
          : ((InstanceProviderTask) serverInstanceProviderTask.getTask())
              .getInstance();
      setResult(builder.bindComponent(null, composite, type, client, clientItf,
          server, serverItf, getCodePieces(), context));
    }
  }

  // --------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // --------------------------------------------------------------------------

  public String[] listFc() {
    return new String[]{BUILDER_ITF};
  }

  public void bindFc(final String itf, final Object value)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException {
    if (itf == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (itf.equals(BUILDER_ITF)) {
      this.builderItf = (BindingBuilder) value;
    } else {
      throw new NoSuchInterfaceException("There is no interface named '" + itf
          + "'");
    }
  }

  public Object lookupFc(final String itf) throws NoSuchInterfaceException {

    if (itf == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (itf.equals(BUILDER_ITF)) {
      return this.builderItf;
    } else {
      throw new NoSuchInterfaceException("There is no interface named '" + itf
          + "'");
    }
  }

  public void unbindFc(final String itf) throws NoSuchInterfaceException,
      IllegalBindingException, IllegalLifeCycleException {

    if (itf == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (itf.equals(BUILDER_ITF)) {
      this.builderItf = null;
    } else {
      throw new NoSuchInterfaceException("There is no interface named '" + itf
          + "'");
    }
  }

}
