/***
 * Cecilia ADL Compiler
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.adl.source;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Map;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.task.codegen.util.SourceFile;

/**
 * Abstract {@link SourceFileBuilder} for builders that returns
 * {@link SourceFile}.
 */
public abstract class AbstractSourceFileBuilder implements SourceFileBuilder {

  // --------------------------------------------------------------------------
  // Implementation of the SourceFileBuilder interface.
  // --------------------------------------------------------------------------

  public SourceFile createSourceFile(final Object componentType,
      final Object componentInstance, final Map<String, Object> codePieces,
      final Map<Object, Object> context) throws Exception {
    final File file = getFile(componentType, componentInstance, context);
    try {
      final PrintStream out = new PrintStream(new FileOutputStream(file));
      out.println("// THIS FILE HAS BEEN GENERATED BY THE CECILIA ADL COMPILER.");
      out.println("// DO NOT EDIT");
      out.println();

      out.println(getFileContent(componentType, componentInstance, codePieces,
          context));
      out.close();
      return getSourceFile(componentType, componentInstance, file);
    } catch (final IOException e) {
      throw new ADLException("An error occurs while writing to file '"
          + file.getAbsolutePath() + "'", null, e);
    }
  }

  /**
   * Returns the {@link SourceFile} object that is created by this builder. This
   * default implementation return a {@link SourceFile} with the
   * <code>componentType</code> as {@link SourceFile#getSignature() signature}.
   * 
   * @see SourceFileBuilder#createSourceFile(Object, Object, Map, Object)
   */
  protected SourceFile getSourceFile(final Object componentType,
      final Object componentInstance, final File file) {
    return new SourceFile(componentType.toString(), file);
  }

  /**
   * Returns the source file content that will be dumped in the file built by
   * this builder.
   * 
   * @see SourceFileBuilder#createSourceFile(Object, Object, Map, Object)
   */
  protected abstract String getFileContent(Object componentType,
      Object componentInstance, Map<String, Object> codePieces,
      Map<Object, Object> context);

  /**
   * Returns the name of the file that is created by this builder. This name if
   * relative to the directory returned by {@link #getOutDir(Object)}.
   * 
   * @see SourceFileBuilder#createSourceFile(Object, Object, Map, Object)
   */
  protected abstract String getFileName(Object componentType,
      Object componentInstance, Map<Object, Object> context);

  /**
   * Utility method to create the file created by this builder.
   * 
   * @see SourceFileBuilder#createSourceFile(Object, Object, Map, Object)
   */
  protected File getFile(final Object componentType,
      final Object componentInstance, final Map<Object, Object> context) {
    final File adlOutDir = getOutDir(context);
    final String fileName = getFileName(componentType, componentInstance,
        context);

    final File dir = (fileName.lastIndexOf(File.separatorChar) == -1)
        ? adlOutDir
        : new File(adlOutDir, fileName.substring(0, fileName
            .lastIndexOf(File.separatorChar)));
    final File file = new File(adlOutDir, fileName);
    dir.mkdirs();
    return file;
  }

  /**
   * Returns the base directory into which this builder will create the source
   * file. The default implementation returns the <code>adl</code>
   * sub-directory of the build directory.
   * 
   * @param context the build context.
   * @return the base directory into which this builder will create the source
   *         file.
   */
  protected File getOutDir(final Map<Object, Object> context) {
    return (File) context.get("adlBuildDirectory");
  }
}